@@ -3,8 +3,17 @@ import ipaddress
 from math import ceil
 from datetime import datetime
 
-from django.core.validators import MinValueValidator, MaxValueValidator
-from django.core.exceptions import ObjectDoesNotExist
+import dns
+from dns import rdata, rdatatype, rdataclass
+
+from django.core.validators import (
+    MinValueValidator,
+    MaxValueValidator,
+    validate_ipv6_address,
+    validate_ipv4_address,
+)
+
+from django.core.exceptions import ObjectDoesNotExist, ValidationError
 from django.db import models, transaction
 from django.db.models import Q, Max, ExpressionWrapper, BooleanField
 from django.db.models.functions import Length
@@ -13,23 +22,23 @@ from django.urls import reverse
 from django.db.models.signals import m2m_changed
 from django.dispatch import receiver
 
-from extras.utils import extras_features
-from netbox.models import PrimaryModel, TaggableManager
 from utilities.querysets import RestrictedQuerySet
+from utilities.choices import ChoiceSet
+
+from netbox.models import NetBoxModel
 
 
-@extras_features("custom_links", "export_templates", "webhooks")
-class NameServer(PrimaryModel):
+def absolute_name(name):
+    if name.endswith("."):
+        return name
+    return f"{name}."
+
+
+class NameServer(NetBoxModel):
     name = models.CharField(
         unique=True,
         max_length=255,
     )
-    tags = TaggableManager(
-        through="extras.TaggedItem",
-        blank=True,
-    )
-
-    objects = RestrictedQuerySet.as_manager()
 
     clone_fields = ["name"]
 
@@ -37,7 +46,7 @@ class NameServer(PrimaryModel):
         ordering = ("name",)
 
     def __str__(self):
-        return self.name
+        return str(self.name)
 
     def get_absolute_url(self):
         return reverse("plugins:netbox_dns:nameserver", kwargs={"pk": self.pk})
@@ -48,7 +57,7 @@ class ZoneManager(models.Manager.from_queryset(RestrictedQuerySet)):
 
     def get_queryset(self):
         return (
-            super(ZoneManager, self)
+            super()
             .get_queryset()
             .annotate(
                 active=ExpressionWrapper(
@@ -58,37 +67,38 @@ class ZoneManager(models.Manager.from_queryset(RestrictedQuerySet)):
         )
 
 
-@extras_features("custom_links", "export_templates", "webhooks")
-class Zone(PrimaryModel):
+class ZoneStatusChoices(ChoiceSet):
+    key = "Zone.status"
+
     STATUS_ACTIVE = "active"
     STATUS_RESERVED = "reserved"
     STATUS_DEPRECATED = "deprecated"
     STATUS_PARKED = "parked"
 
-    CHOICES = (
-        (STATUS_ACTIVE, "Active"),
-        (STATUS_RESERVED, "Reserved"),
-        (STATUS_DEPRECATED, "Deprecated"),
-        (STATUS_PARKED, "Parked"),
+    CHOICES = [
+        (STATUS_ACTIVE, "Active", "blue"),
+        (STATUS_RESERVED, "Reserved", "cyan"),
+        (STATUS_DEPRECATED, "Deprecated", "red"),
+        (STATUS_PARKED, "Parked", "gray"),
+    ]
+
+
+class Zone(NetBoxModel):
+    ACTIVE_STATUS_LIST = (ZoneStatusChoices.STATUS_ACTIVE,)
+
+    view = models.ForeignKey(
+        to="View",
+        on_delete=models.PROTECT,
+        blank=True,
+        null=True,
     )
-
-    CSS_CLASSES = {
-        STATUS_ACTIVE: "primary",
-        STATUS_RESERVED: "info",
-        STATUS_DEPRECATED: "danger",
-        STATUS_PARKED: "warning",
-    }
-
-    ACTIVE_STATUS_LIST = (STATUS_ACTIVE,)
-
     name = models.CharField(
-        unique=True,
         max_length=255,
     )
     status = models.CharField(
         max_length=50,
-        choices=CHOICES,
-        default=STATUS_ACTIVE,
+        choices=ZoneStatusChoices,
+        default=ZoneStatusChoices.STATUS_ACTIVE,
         blank=True,
     )
     nameservers = models.ManyToManyField(
@@ -96,10 +106,6 @@ class Zone(PrimaryModel):
         related_name="zones",
         blank=True,
     )
-    tags = TaggableManager(
-        through="extras.TaggedItem",
-        blank=True,
-    )
     default_ttl = models.PositiveIntegerField(
         blank=True,
         verbose_name="Default TTL",
@@ -164,6 +170,7 @@ class Zone(PrimaryModel):
     objects = ZoneManager()
 
     clone_fields = [
+        "view",
         "name",
         "status",
         "nameservers",
@@ -178,10 +185,23 @@ class Zone(PrimaryModel):
     ]
 
     class Meta:
-        ordering = ("name",)
+        ordering = (
+            "view",
+            "name",
+        )
+        unique_together = (
+            "view",
+            "name",
+        )
 
     def __str__(self):
-        return self.name
+        if self.view:
+            return f"[{self.view}] {self.name}"
+
+        return str(self.name)
+
+    def get_status_color(self):
+        return ZoneStatusChoices.colors.get(self.status)
 
     def get_absolute_url(self):
         return reverse("plugins:netbox_dns:zone", kwargs={"pk": self.pk})
@@ -193,12 +213,15 @@ class Zone(PrimaryModel):
         soa_name = "@"
         soa_ttl = self.soa_ttl
         soa_value = (
-            f"({self.soa_mname} {self.soa_rname} {self.soa_serial}"
+            f"({absolute_name(self.soa_mname.name)} {absolute_name(self.soa_rname)}"
+            f" {self.soa_serial}"
             f" {self.soa_refresh} {self.soa_retry} {self.soa_expire}"
             f" {self.soa_minimum})"
         )
 
-        old_soa_records = self.record_set.filter(type=Record.SOA, name=soa_name)
+        old_soa_records = self.record_set.filter(
+            type=RecordTypeChoices.SOA, name=soa_name
+        )
 
         if len(old_soa_records):
             for index, record in enumerate(old_soa_records):
@@ -214,7 +237,7 @@ class Zone(PrimaryModel):
         else:
             Record.objects.create(
                 zone_id=self.pk,
-                type=Record.SOA,
+                type=RecordTypeChoices.SOA,
                 name=soa_name,
                 ttl=soa_ttl,
                 value=soa_value,
@@ -225,16 +248,16 @@ class Zone(PrimaryModel):
         ns_name = "@"
         ns_ttl = self.default_ttl
 
-        delete_ns = self.record_set.filter(type=Record.NS, managed=True).exclude(
-            value__in=nameservers
-        )
+        delete_ns = self.record_set.filter(
+            type=RecordTypeChoices.NS, managed=True
+        ).exclude(value__in=nameservers)
         for record in delete_ns:
             record.delete()
 
         for ns in nameservers:
             Record.raw_objects.update_or_create(
                 zone_id=self.pk,
-                type=Record.NS,
+                type=RecordTypeChoices.NS,
                 name=ns_name,
                 ttl=ns_ttl,
                 value=ns,
@@ -248,15 +271,19 @@ class Zone(PrimaryModel):
         ns_errors = []
 
         if not nameservers:
-            ns_errors.append(f"No nameservers are configured for zone {self.name}")
+            ns_errors.append(f"No nameservers are configured for zone {self}")
 
         for nameserver in nameservers:
             ns_domain = ".".join(nameserver.name.split(".")[1:])
             if not ns_domain:
                 continue
 
+            view_condition = (
+                Q(view__isnull=True) if self.view is None else Q(view_id=self.view.pk)
+            )
+
             try:
-                ns_zone = Zone.objects.get(name=ns_domain)
+                ns_zone = Zone.objects.get(view_condition, name=ns_domain)
             except ObjectDoesNotExist:
                 continue
 
@@ -264,18 +291,18 @@ class Zone(PrimaryModel):
             address_records = Record.objects.filter(
                 Q(zone=ns_zone),
                 Q(Q(name=f"{nameserver.name}.") | Q(name=ns_name)),
-                Q(Q(type=Record.A) | Q(type=Record.AAAA)),
+                Q(Q(type=RecordTypeChoices.A) | Q(type=RecordTypeChoices.AAAA)),
             )
 
             if not address_records:
                 ns_warnings.append(
-                    f"Nameserver {nameserver.name} does not have an address record in zone {ns_zone.name}"
+                    f"Nameserver {nameserver.name} does not have an address record in zone {ns_zone}"
                 )
 
         return ns_warnings, ns_errors
 
     def get_auto_serial(self):
-        records = Record.objects.filter(zone=self).exclude(type=Record.SOA)
+        records = Record.objects.filter(zone=self).exclude(type=RecordTypeChoices.SOA)
         if records:
             soa_serial = (
                 records.aggregate(Max("last_updated"))
@@ -300,30 +327,64 @@ class Zone(PrimaryModel):
             f'{".".join(zone_fields[length:])}' for length in range(1, len(zone_fields))
         ]
 
+    def check_name_conflict(self):
+        if self.view is None:
+            if (
+                Zone.objects.exclude(pk=self.pk)
+                .filter(name=self.name, view__isnull=True)
+                .exists()
+            ):
+                raise ValidationError(
+                    {"name": "A zone with name %(name)s and no view already exists."},
+                    params={"name": self.name},
+                )
+
+    def clean(self, *args, **kwargs):
+        self.check_name_conflict()
+
+        if self.pk is None:
+            return
+
+        old_zone = Zone.objects.get(pk=self.pk)
+        if old_zone.view == self.view and old_zone.status == self.status:
+            return
+
     def save(self, *args, **kwargs):
+        self.check_name_conflict()
+
         new_zone = self.pk is None
         if not new_zone:
-            renamed_zone = Zone.objects.get(pk=self.pk).name != self.name
+            old_zone = Zone.objects.get(pk=self.pk)
+            renamed_zone = old_zone.name != self.name
+            changed_view = old_zone.view != self.view
+            changed_status = old_zone.status != self.status
         else:
             renamed_zone = False
+            changed_view = False
+            changed_status = False
 
         if self.soa_serial_auto:
             self.soa_serial = self.get_auto_serial()
 
         super().save(*args, **kwargs)
 
-        if (new_zone or renamed_zone) and self.name.endswith(".arpa"):
+        if (
+            new_zone or renamed_zone or changed_view or changed_status
+        ) and self.name.endswith(".arpa"):
             address_records = Record.objects.filter(
                 Q(ptr_record__isnull=True)
-                | Q(ptr_record__zone__name__in=self.parent_zones()),
-                type__in=(Record.A, Record.AAAA),
+                | Q(ptr_record__zone__name__in=self.parent_zones())
+                | Q(ptr_record__zone__name=self.name),
+                type__in=(RecordTypeChoices.A, RecordTypeChoices.AAAA),
                 disable_ptr=False,
             )
             for record in address_records:
                 record.update_ptr_record()
 
-        elif renamed_zone:
-            for record in self.record_set.filter(ptr_record__isnull=False):
+        elif renamed_zone or changed_view or changed_status:
+            for record in self.record_set.filter(
+                type__in=(RecordTypeChoices.A, RecordTypeChoices.AAAA)
+            ):
                 record.update_ptr_record()
 
         self.update_soa_record()
@@ -364,7 +425,7 @@ class RecordManager(models.Manager.from_queryset(RestrictedQuerySet)):
 
     def get_queryset(self):
         return (
-            super(RecordManager, self)
+            super()
             .get_queryset()
             .annotate(
                 active=ExpressionWrapper(
@@ -381,79 +442,47 @@ class RecordManager(models.Manager.from_queryset(RestrictedQuerySet)):
                 )
             )
         )
-        return queryset
 
 
-@extras_features("custom_links", "export_templates", "webhooks")
-class Record(PrimaryModel):
-    A = "A"
-    AAAA = "AAAA"
-    CNAME = "CNAME"
-    MX = "MX"
-    TXT = "TXT"
-    NS = "NS"
-    SOA = "SOA"
-    SRV = "SRV"
-    PTR = "PTR"
-    SPF = "SPF"
-    CAA = "CAA"
-    DS = "DS"
-    SSHFP = "SSHFP"
-    TLSA = "TLSA"
-    AFSDB = "AFSDB"
-    APL = "APL"
-    DNSKEY = "DNSKEY"
-    CDNSKEY = "CDNSKEY"
-    CERT = "CERT"
-    DCHID = "DCHID"
-    DNAME = "DNAME"
-    HIP = "HIP"
-    IPSECKEY = "IPSECKEY"
-    LOC = "LOC"
-    NAPTR = "NAPTR"
-    NSEC = "NSEC"
-    RRSIG = "RRSIG"
-    RP = "RP"
+def initialize_choice_names(cls):
+    for choice in cls.CHOICES:
+        setattr(cls, choice[0], choice[0])
+    return cls
 
-    CHOICES = (
-        (A, A),
-        (AAAA, AAAA),
-        (CNAME, CNAME),
-        (MX, MX),
-        (TXT, TXT),
-        (SOA, SOA),
-        (NS, NS),
-        (SRV, SRV),
-        (PTR, PTR),
-        (SPF, SPF),
-        (CAA, CAA),
-        (DS, DS),
-        (SSHFP, SSHFP),
-        (TLSA, TLSA),
-        (AFSDB, AFSDB),
-        (APL, APL),
-        (DNSKEY, DNSKEY),
-        (CDNSKEY, CDNSKEY),
-        (CERT, CERT),
-        (DCHID, DCHID),
-        (DNAME, DNAME),
-        (HIP, HIP),
-        (IPSECKEY, IPSECKEY),
-        (LOC, LOC),
-        (NAPTR, NAPTR),
-        (NSEC, NSEC),
-        (RRSIG, RRSIG),
-        (RP, RP),
+
+@initialize_choice_names
+class RecordTypeChoices(ChoiceSet):
+    CHOICES = [
+        (rdtype.name, rdtype.name)
+        for rdtype in sorted(rdatatype.RdataType, key=lambda a: a.name)
+        if not rdatatype.is_metatype(rdtype)
+    ]
+    SINGLETONS = [
+        rdtype.name for rdtype in rdatatype.RdataType if rdatatype.is_singleton(rdtype)
+    ]
+
+
+@initialize_choice_names
+class RecordClassChoices(ChoiceSet):
+    CHOICES = [
+        (rdclass.name, rdclass.name)
+        for rdclass in sorted(rdataclass.RdataClass)
+        if not rdataclass.is_metaclass(rdclass)
+    ]
+
+
+class Record(NetBoxModel):
+    unique_ptr_qs = Q(
+        Q(disable_ptr=False),
+        Q(Q(type=RecordTypeChoices.A) | Q(type=RecordTypeChoices.AAAA)),
     )
 
-    unique_ptr_qs = Q(Q(disable_ptr=False), Q(Q(type="A") | Q(type="AAAA")))
-
     zone = models.ForeignKey(
         Zone,
         on_delete=models.CASCADE,
     )
     type = models.CharField(
-        choices=CHOICES,
+        choices=RecordTypeChoices,
         max_length=10,
     )
     name = models.CharField(
@@ -465,10 +494,6 @@ class Record(PrimaryModel):
     ttl = models.PositiveIntegerField(
         verbose_name="TTL",
     )
-    tags = TaggableManager(
-        through="extras.TaggedItem",
-        blank=True,
-    )
     managed = models.BooleanField(
         null=False,
         default=False,
@@ -494,24 +519,15 @@ class Record(PrimaryModel):
 
     class Meta:
         ordering = ("zone", "name", "type", "value")
-        constraints = (
-            models.UniqueConstraint(
-                name="unique_pointer_for_address",
-                fields=["type", "value"],
-                condition=(
-                    models.Q(
-                        models.Q(disable_ptr=False),
-                        models.Q(type="A") | models.Q(type="AAAA"),
-                    )
-                ),
-            ),
-        )
 
     def __str__(self):
+        if self.name == "@":
+            return f"{self.zone.name} [{self.type}]"
+
         if self.name.endswith("."):
             return f"{self.name} [{self.type}]"
-        else:
-            return f"{self.name}.{self.zone.name} [{self.type}]"
+
+        return f"{self.name}.{self.zone.name} [{self.type}]"
 
     def get_absolute_url(self):
         return reverse("plugins:netbox_dns:record", kwargs={"pk": self.id})
@@ -519,7 +535,7 @@ class Record(PrimaryModel):
     def fqdn(self):
         return f"{self.name}.{self.zone.name}."
 
-    def ptr_zone(self):
+    def ptr_zone(self, view=None):
         address = ipaddress.ip_address(self.value)
         if address.version == 4:
             lengths = range(1, 4)
@@ -530,12 +546,20 @@ class Record(PrimaryModel):
             ".".join(address.reverse_pointer.split(".")[length:]) for length in lengths
         ]
 
-        ptr_zones = Zone.objects.filter(Q(name__in=zone_names)).order_by(
-            Length("name").desc()
-        )
+        if view is None:
+            view = self.zone.view
+
+        if view is None:
+            ptr_zone_filter = Q(name__in=zone_names, view__isnull=True)
+        else:
+            ptr_zone_filter = Q(name__in=zone_names, view_id=view.pk)
+
+        ptr_zones = Zone.objects.filter(ptr_zone_filter).order_by(Length("name").desc())
         if len(ptr_zones):
             return ptr_zones[0]
 
+        return None
+
     def update_ptr_record(self):
         ptr_zone = self.ptr_zone()
 
@@ -572,7 +596,7 @@ class Record(PrimaryModel):
             if ptr_record is None:
                 ptr_record = Record.objects.create(
                     zone_id=ptr_zone.pk,
-                    type=Record.PTR,
+                    type=RecordTypeChoices.PTR,
                     name=ptr_name,
                     ttl=self.ttl,
                     value=ptr_value,
@@ -582,14 +606,78 @@ class Record(PrimaryModel):
         self.ptr_record = ptr_record
         super().save()
 
+    def clean_fields(self, *args, **kwargs):
+        self.type = self.type.upper()
+        super().clean_fields(*args, **kwargs)
+
+    def clean(self, *args, **kwargs):
+        ip_version = None
+
+        if self.ttl is None:
+            self.ttl = self.zone.default_ttl
+
+        try:
+            if self.type == RecordTypeChoices.A:
+                ip_version = "4"
+                validate_ipv4_address(self.value)
+            elif self.type == RecordTypeChoices.AAAA:
+                ip_version = "6"
+                validate_ipv6_address(self.value)
+            else:
+                rdata.from_text(RecordClassChoices.IN, self.type, self.value)
+
+        except ValidationError:
+            raise ValidationError(
+                {
+                    "value": f"A valid IPv{ip_version} address is required for record type {self.type}."
+                }
+            ) from None
+
+        except dns.exception.SyntaxError as exc:
+            raise ValidationError(
+                {"value": f"Record value {self.value} is malformed: {exc}."}
+            ) from None
+
+        records = Record.objects.filter(name=self.name, zone=self.zone).exclude(
+            pk=self.pk
+        )
+
+        if self.type == RecordTypeChoices.CNAME:
+            if records.exists():
+                raise ValidationError(
+                    {
+                        "type": f"There is already a record for name {self.name} in zone {self.zone}, CNAME is not allowed."
+                    }
+                ) from None
+
+        elif records.filter(type=RecordTypeChoices.CNAME).exists():
+            raise ValidationError(
+                {
+                    "type": f"There is already a CNAME record for name {self.name} in zone {self.zone}, no other record allowed."
+                }
+            ) from None
+
+        elif self.type in RecordTypeChoices.SINGLETONS:
+            if records.filter(type=self.type).exists():
+                raise ValidationError(
+                    {
+                        "type": f"There is already a {self.type} record for name {self.name} in zone {self.zone}, more than one are not allowed."
+                    }
+                ) from None
+
     def save(self, *args, **kwargs):
-        if self.type in (self.A, self.AAAA):
+        self.full_clean()
+
+        if self.type in (RecordTypeChoices.A, RecordTypeChoices.AAAA):
             self.update_ptr_record()
+        elif self.ptr_record is not None:
+            self.ptr_record.delete()
+            self.ptr_record = None
 
         super().save(*args, **kwargs)
 
         zone = self.zone
-        if self.type != self.SOA and zone.soa_serial_auto:
+        if self.type != RecordTypeChoices.SOA and zone.soa_serial_auto:
             zone.update_serial()
 
     def delete(self, *args, **kwargs):
@@ -601,3 +689,21 @@ class Record(PrimaryModel):
         zone = self.zone
         if zone.soa_serial_auto:
             zone.update_serial()
+
+
+class View(NetBoxModel):
+    name = models.CharField(
+        unique=True,
+        max_length=255,
+    )
+
+    clone_fields = ["name"]
+
+    def get_absolute_url(self):
+        return reverse("plugins:netbox_dns:view", kwargs={"pk": self.id})
+
+    def __str__(self):
+        return str(self.name)
+
+    class Meta:
+        ordering = ("name",)
